#!/usr/bin/env python3

from . mqtt import MqttClient, MQTTv311, topic
from functools import partial, wraps
import io, json, logging, toml, uuid, wave


__all__ = ('SnipsClient', 'debug_json',
    'on_hotword_detected', 'on_start_session', 'on_session_started', 'on_intent',
    'on_end_session', 'on_continue_session', 'on_session_ended', 'on_play_finished' )


class SnipsClient(MqttClient):
    'Snips client with auto-configuration'
    
    CONFIG = '/etc/snips.toml'
    
    INTENT_PREFIX         = 'hermes/intent/'
    DIALOGUE              = 'hermes/dialogueManager/'
    
    # Session life cycle messages
    HOTWORD_DETECTED      = 'hermes/hotword/+/detected'
    START_SESSION         = DIALOGUE + 'startSession'
    SESSION_QUEUED        = DIALOGUE + 'sessionQueued'
    SESSION_STARTED       = DIALOGUE + 'sessionStarted'
    CONTINUE_SESSION      = DIALOGUE + 'continueSession'
    END_SESSION           = DIALOGUE + 'endSession'
    SESSION_ENDED         = DIALOGUE + 'sessionEnded'
    
    # Misc
    PLAY_BYTES     = 'hermes/audioServer/{site_id}/playBytes/{request_id}'
    PLAY_FINISHED  = 'hermes/audioServer/%s/playFinished'
    REGISTER_SOUND = 'hermes/tts/registerSound/%s'


    def __init__(self, config=CONFIG, 
        client_id=None, clean_session=True, userdata=None,
        protocol=MQTTv311, transport=MqttClient.TCP):
        
        super().__init__(client_id, clean_session, userdata,
            protocol, transport)

        self.log.debug('Loading config: %s', config)        
        self.config = toml.load(config)

        
    def connect(self):
        'Connect to the MQTT broker and invoke callback methods'
        common = self.config.get('snips-common', {})
        
        host, port = None, None
        host_port = common.get('mqtt', 'localhost:1883')
        if host_port:
            if ':' in host_port:
                host, port = host_port.split(':')
                port = int(port)
            else:
                host = host_port
        
        password = None
        username = common.get('mqtt_username')
        if username:
            password = common.get('mqtt_password')
        
        cafile = common.get('mqtt_tls_cafile')
        cert = common.get('mqtt_tls_client_cert')
        key = None if not cert else common.get('mqtt_tls_client_key')
        
        if cafile or cert or port == self.DEFAULT_TLS_PORT:
            assert not common.get('mqtt_tls_hostname'), \
                'mqtt_tls_hostname not supported'
            self.tls_set(ca_certs=cafile, certfile=cert, keyfile=key)
            self._tls_initialized = True
        
        return super().connect(host=host, port=port,
            username=username, password=password)


    # See: https://docs.snips.ai/reference/dialogue#session-initialization-action
    def action_init(self, text=None, intent_filter=[],
            can_be_enqueued=True, send_intent_not_recognized=False):
        'Build the init part of action type to start a session'
        
        init = { 'type' : 'action' }
        if text: init['text'] = str(text)
        if not can_be_enqueued: init['canBeEnqueued'] = False
        if intent_filter: init['intentFilter'] = intent_filter
        if send_intent_not_recognized: init['sendIntentNotRecognized'] = True
        return init


    # See: https://docs.snips.ai/reference/dialogue#session-initialization-notification
    def notification_init(self, text):
        'Build the init part of notification type to start a session'
        return { 'type' : 'notification', 'text' : str(text) }


    # See: https://docs.snips.ai/reference/dialogue#start-session
    def start_session(self, site_id, init, custom_data=None, qos=1):
        'End the session with an optional message'
        payload = { 'siteId': site_id, 'init' : init }
        
        if type(custom_data) in (dict, list, tuple):
            payload['customData'] = json.dumps(custom_data)
        elif custom_data is not None:
            payload['customData'] = str(custom_data)
            
        self.log.debug("Starting %s session on site '%s'", init.get('type'), site_id)
        self.publish(self.START_SESSION, json.dumps(payload), qos=qos)


    def speak(self, site_id, text):
        'Say a one-time notification'
        self.start_session(site_id, self.notification_init(text))
        

    # See: https://docs.snips.ai/reference/dialogue#end-session
    def end_session(self, session_id, text=None, qos=1):
        'End the session with an optional message'
        payload = { 'sessionId': session_id }
        
        if text:
            text = ' '.join(text.split())
            payload['text'] = text

        self.log.debug("Ending session %s with '%s'", session_id, text)
        self.publish(self.END_SESSION, json.dumps(payload), qos=qos)


    # See: https://docs.snips.ai/reference/dialogue#continue-session
    def continue_session(self, session_id, text, intent_filter=None, slot=None,
            send_intent_not_recognized=False, custom_data=None, qos=1):
        'Continue the session with a question'
        
        text = ' '.join(text.split())
        payload = { 'text': text, 'sessionId': session_id }
        if intent_filter: payload['intentFilter'] = intent_filter
        if slot: payload['slot'] = slot
        if send_intent_not_recognized:
            payload['sendIntentNotRecognized'] = bool(send_intent_not_recognized)

        if type(custom_data) in (dict, list, tuple):
            payload['customData'] = json.dumps(custom_data)
        elif custom_data is not None:
            payload['customData'] = str(custom_data)

        self.log.debug("Continuing session %s with '%s'", session_id, text)
        self.publish(self.CONTINUE_SESSION, json.dumps(payload), qos=qos)


    # See: https://docs.snips.ai/reference/dialogue#start-session
    def play_sound(self, site_id, wav_data, request_id=None):
        'Play a WAV sound at the given site'
        if not request_id: request_id = str(uuid.uuid4())
        self.publish(self.PLAY_BYTES.format(site_id=site_id, request_id=request_id), payload=wav_data)
        return request_id


    def register_sound(self, name, wav_data):
        self.publish(self.REGISTER_SOUND % name, wav_data)
        return self


    def run(self):
        'Connect to MQTT and handle incoming messages'
        with self.connect():
            self.loop_forever()


###################################
### Decorators for Snips events ###
###################################

def _load_json(payload):
    'Helper to convert JSON to a Python dict'
    # Only convert if this appears to be a JSON payload.
    # Needed for multiple annotations on a method
    return json.loads(payload) if type(payload) is bytes else payload
    
on_hotword_detected = partial(topic,
    SnipsClient.HOTWORD_DETECTED, payload_converter=_load_json)

on_start_session = partial(topic,
    SnipsClient.START_SESSION, payload_converter=_load_json)

def on_intent(intent, qos=0, log_level=logging.DEBUG):
 return topic('%s%s' % (SnipsClient.INTENT_PREFIX, intent),
     qos=qos, payload_converter=_load_json, log_level=log_level)

on_continue_session = partial(topic,
    SnipsClient.CONTINUE_SESSION, payload_converter=_load_json)

on_session_queued = partial(topic,
    SnipsClient.SESSION_QUEUED, payload_converter=_load_json)

on_session_started = partial(topic,
    SnipsClient.SESSION_STARTED, payload_converter=_load_json)

on_end_session = partial(topic,
    SnipsClient.END_SESSION, payload_converter=_load_json)

on_session_ended = partial(topic,
    SnipsClient.SESSION_ENDED, payload_converter=_load_json)

def on_play_finished(site='+', qos=0, log_level=logging.DEBUG):
    return topic(SnipsClient.PLAY_FINISHED % site,
        qos=qos, payload_converter=_load_json, log_level=log_level)


def debug_json(keys=[]):
    'Decorator to debug message payloads'
    def wrapper(method):
        @wraps(method)
        def wrapped(client, userdata, msg):
            if type(msg.payload) is dict:
                data = msg.payload
                if keys: data = { k: v for k, v in data.items()
                    if not keys or k in keys }
                client.log.debug('Payload: %s',
                    json.dumps(data, sort_keys=True, indent=2))
            return method(client, userdata, msg)
        return wrapped
    return wrapper



from . exceptions import SnipsError, SnipsClarificationError
from . i18n import get_translations
from . intent import IntentPayload
from . logging import LoggingMixin
from . mqtt import CommandLineMixin
from . snips import SnipsClient, on_intent
from configparser import ConfigParser
from functools import wraps
import logging, os


__all__ = ('Skill', 'intent', 'min_confidence', 'PARDON', 'require_slot' )
_, ngettext = get_translations(__file__, 'snips_skill')


PARDON = _('Pardon?')


class Skill(LoggingMixin, CommandLineMixin, SnipsClient):
    'Base class for Snips actions.'
    
    CONFIGURATION_FILE = 'config.ini'
    
    STANDARD_SECTIONS = ('DEFAULT', 'global', 'secret')
    
    
    def __init__(self):
        super().__init__()
        self.configuration = ConfigParser()

        if os.path.isfile(self.options.config):
            self.log.debug('Reading configuration: %s', self.options.config)
            self.configuration.read(self.options.config, encoding='UTF-8')
            self.process_config()
        else:
            self.log.warning('Configuration %s not found', self.options.config)

        
    def process_config(self):
        'May be overridden'
        pass


    def get_config(self, section='DEFAULT'):
        'Get a configuration section, or DEFAULT values'
        if section in self.configuration:
            return self.configuration[section]
        return self.configuration['DEFAULT']


    def add_arguments(self):
        super().add_arguments()
        self.parser.add_argument('-c', '--config',
            default=self.CONFIGURATION_FILE,
            help='Configuration file (%s)' % self.CONFIGURATION_FILE)
    
    
def intent(intent, qos=1, log_level=logging.DEBUG, silent=False):
    ''' Decorator for intent handlers.
        :param intent: Intent name.
        :param qos: MQTT quality of service.
        :param log_level: Log intents at this level, if set.
        :param silent: Set to `True` for intents that should return `None` 
        The wrapped function gets a parsed `IntentPayload` object
        instead of a JSON `msg.payload`.
        If a `SnipsClarificationError` is raised, the session continues with a question.
        Otherwise, the session is ended.
    '''

    def wrapper(method):
        @on_intent(intent, qos=qos, log_level=logging.DEBUG)
        @wraps(method)
        def wrapped(client, userdata, msg):
            msg.payload = IntentPayload(msg.payload)
            if log_level: client.log_intent(msg.payload, level=log_level)
            try:
                result = method(client, userdata, msg)
                if result is None and silent:
                    client.end_session(msg.payload.session_id, qos=qos)
                    return
                raise SnipsError(result)
                if log_level: client.log_response(result, level=log_level)
            except SnipsClarificationError as sce:
                if log_level: client.log_response(sce, level=log_level)
                client.continue_session(msg.payload.session_id, str(sce),
                    [sce.intent ] if sce.intent else None,
                    slot=sce.slot, custom_data=sce.custom_data)
            except SnipsError as e:
                if log_level: client.log_response(e, level=log_level)
                client.end_session(msg.payload.session_id, str(e), qos=qos)
        return wrapped
    return wrapper


def min_confidence(threshold, prompt=PARDON):
    ''' Decorator that requires a minimum intent confidence, or else asks the user.
        :param threshold: Minimum confidence (0.0 - 1.0)
        :param prompt: Question for the user
    '''

    def wrapper(method):
        @wraps(method)
        def wrapped(client, userdata, msg):
            if msg.payload.intent.confidence_score >= threshold:
                return method(client, userdata, msg)
            raise SnipsClarificationError(prompt)
        return wrapped
    return wrapper


def require_slot(slot, prompt, kind=None):
    ''' Decorator that checks whether a slot is present, or else asks the user.
        :param slot: Slot name
        :param prompt: Question for the user
        :param kind: Optionally, the slot is expected to be of the given kind.
    '''

    def wrapper(method):
        @wraps(method)
        def wrapped(client, userdata, msg):
            if slot in msg.payload.slots and (kind is None
                or msg.payload.slot_values[slot].kind == kind):
                    return method(client, userdata, msg)
            raise SnipsClarificationError(prompt,
                msg.payload.intent.intent_name, slot)
        return wrapped
    return wrapper



from . mqtt import topic
import re


__all__ = ('StateAwareMixin',)


class StateAwareMixin:
    ''' Mixin for stateful skills.
        Status updates are recorded in-memory from MQTT topics,
        e.g. `status/#`.
        The message payload for status updates is JSON-converted if possible.
        The last known state is available in `self.current_state`.
        Subclasses may define handler methods for particular topics,
        e.g. `on_status_lamp_brightness(payload)`.
    '''
    
    replacements = str.maketrans('/-', '__')
    wildcard_map = { '+' : r'([^/]+)', '#' : r'(.*)' }

    def __init__(self):
        'Register topics and the state callcack.'
        
        super().__init__()
        self.current_state = {}

        status_topic = self.get_config().get('status_topic')
        assert status_topic, 'status_topic not found in configuration'

        # Construct topic RE from status topic
        self.topic_pattern = re.compile('/'.join(
            map(lambda part: self.wildcard_map.get(part, part),
                status_topic.split('/'))))
        
        # Subscribe to status updates
        register = topic(status_topic, payload_converter=self.decode_json)
        register(self.update_status)


    @staticmethod
    def update_status(self, _userdata, msg):
        ''' Track the global state,
            and invoke handler methods defined by subclasses
            with the message payload.
        '''
        self.record_state(msg.topic, msg.payload)
        self.on_status_update(msg.topic, msg.payload)


    def record_state(self, topic, payload):
        'Keep the global state in-memory'
        store = self.current_state
        keys = self.split_topic(topic)
        for key in keys[:-1]:
            if key not in store: store[key] = {}
            store = store[key]
        store[keys[-1]] = payload
        self.log.debug('Updated: %s = %s', '.'.join(keys), payload)
        
        
    def split_topic(self, topic):
        ''' Extract relevant components from a message topic.
            The default implementation extracts all parts of the 
            topic corresponding to a MQTT topic wildcard.
            Subclasses may override this.
        '''
        match = self.topic_pattern.match(topic)
        assert match, 'topic pattern mismatch'
        return [part for group in match.groups() for part in group.split('/')]

        
    def on_status_update(self, topic, payload):
        ''' Generic handler for status updates,
            tries to find and invoke specific handlers.
        '''
        method_name = 'on_%s' % topic.translate(self.replacements)
        if hasattr(self, method_name):
            handler = getattr(self, method_name)
            self.log.debug('Invoking %s', method_name)
            handler(payload)



from . intent import IntentPayload
from . logging import LoggingMixin
from . mqtt import CommandLineMixin
from . snips import *
import logging


if __name__ == '__main__': # demo code

    class Logger(CommandLineMixin, LoggingMixin, SnipsClient):
            
        @on_intent('#')
        def intent_logger(self, userdata, msg):
            self.log_intent(IntentPayload(msg.payload), level=logging.INFO)

        @on_end_session(log_level=None)
        @on_continue_session()
        def response_logger(self, userdata, msg):
            self.log_response(msg.payload.get('text'), level=logging.INFO)

    Logger().run()



class SnipsError(Exception):
    'Signal that an intent cannot be handled'
    
    def __str__(self):
        return str(self.args[0])


class SnipsClarificationError(SnipsError):
    'Signal that some clarification is needed from the user.'
    
    def __init__(self, msg, intent=None, slot=None, custom_data=None):
        ''' Create a clarification request.
            :param msg: Question for the user
            :param intent: Optional expected intent
            :param slot: Optional slot name
            See: https://docs.snips.ai/reference/dialogue#continue-session
        '''
        super().__init__(msg, intent, slot, custom_data)
        
    def __str__(self):
        return str(self.args[0])
        
    @property
    def intent(self):
        return self.args[1]

    @property
    def slot(self):
        return self.args[2]
        
    @property
    def custom_data(self):
        return self.args[3]



from colors import red, green, yellow, blue, magenta, cyan
from functools import partial
import logging, re, sys

__all__ = ('LoggingMixin', )


class LoggingMixin:
    'Logging for Snips events'
    
    INDENT = 10
    STRIP_COLOR = partial(re.compile('\x1b\\[(K|.*?m)').sub, '')
    COLORIZE = str if sys.stdout.isatty() else STRIP_COLOR


    def colored_log(self, level, format, *args, color=str):
        self.log.log(level, format, *map(self.COLORIZE,
            map(color, args)))
    

    def tabular_log(self, level, key, value, label_color=str, width=INDENT):
        label = label_color('%-*s') % (width, key)
        self.colored_log(level, '%s %s', label, str(value))
    
    
    def log_intent(self, payload, level=logging.DEBUG):
        'Log an intent message'
        self.tabular_log(level, 'intent', '%s, confidence: %.1f' % (
            red(payload.intent.intent_name, style='bold'),
            payload.intent.confidence_score), label_color=green)
        for k in ('site_id', 'input'):
            self.tabular_log(level, k, getattr(payload, k), label_color=cyan)
        for name, slot in payload.slots.items():
            self.tabular_log(level, name, slot.value, label_color=magenta)
        if payload.custom_data:
            self.tabular_log(level, 'data', payload.custom_data, label_color=yellow)
            
    
    def log_response(self, response, level=logging.DEBUG):
        'Log an action response'
        if response: self.tabular_log(level, 'answer',
            red(response, style='bold'), label_color=green)



#!/usr/bin/env python3

import json
from datetime import datetime, timedelta


__all__ = ('IntentPayload',)


def parse_date(date_str):
    'Convert a Snips date string to datetime'
    date_str = date_str[:19] + date_str[-7:-3] + date_str[-2:]
    return datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S %z")


class SlotValue:

    def __init__(self, json_dict):
        self.kind = json_dict['kind']
        self.value = json_dict.get('value')
    
    def __repr__(self):
        return "<%s value: %s>" % (self.kind, self.value)
        

# See: https://docs.snips.ai/articles/platform/dialog/slot-types#time-related-entities
class InstantTimeValue(SlotValue):

    def __init__(self, json_dict):
        super().__init__(json_dict)
        self.grain = json_dict['grain']
        self.precision = json_dict['precision']
        self.value = parse_date(self.value)
        

class TimeIntervalValue(SlotValue):

    def __init__(self, json_dict):
        super().__init__(json_dict)
        t1, t2 = json_dict['from'], json_dict['to']
        self.value = (parse_date(t1) if t1 else None, parse_date(t2) if t2 else None)
        

# See: https://docs.snips.ai/articles/platform/dialog/slot-types#duration-entities
# TODO: years, quarters and months are not parsed
class DurationValue(SlotValue):

    def __init__(self, json_dict):
        super().__init__(json_dict)
        self.precision = json_dict['precision']
        self.value = timedelta(
            weeks=json_dict['weeks'],
            days=json_dict['days'],
            hours=json_dict['hours'],
            minutes=json_dict['minutes'],
            seconds=json_dict['seconds'])
        

class TemperatureValue(SlotValue):

    def __init__(self, json_dict):
        super().__init__(json_dict)
        self.unit = json_dict['unit']
    
    def __repr__(self):
        return "<%s value: %s %s>" % (self.kind, self.value, self.unit)
        

class MonetaryValue(SlotValue):

    def __init__(self, json_dict):
        super().__init__(json_dict)
        self.unit = json_dict['unit']
        self.precision = json_dict['precision']
    
    def __repr__(self):
        return "<%s value: %s %s>" % (self.kind, self.value, self.unit)
        

# See: https://docs.snips.ai/reference/dialogue#slot
class Slot:

    VALUE_MAP = {
        'InstantTime'   : InstantTimeValue,
        'TimeInterval'  : TimeIntervalValue,
        'Duration'      : DurationValue,
        'Temperature'   : TemperatureValue,
        'AmountOfMoney' : MonetaryValue,
    }
    
    
    def __init__(self, position, json_dict):
        self.position = position
        self.entity = json_dict['entity']
        self.slot_name = json_dict['slotName']
        self.raw_value = json_dict['rawValue']
        self.confidence_score = json_dict['confidenceScore']
        
        r = json_dict['range']
        self.range = (r['start'], r['end']) if r else None

        value = json_dict['value']
        cls = self.VALUE_MAP.get(value['kind'], SlotValue)
        self.value = cls(value)


    def __repr__(self):
        return "<Slot '%s' score=%.2f value='%s'>" % (
            self.slot_name, self.confidence_score, self.value)


# See: https://docs.snips.ai/reference/dialogue#intent
class Intent:
    
    def __init__(self, json_dict):
        self.intent_name = json_dict['intentName']
        self.confidence_score = json_dict['confidenceScore']

    def __repr__(self):
        return "<Intent '%s' score=%.2f>" % (
            self.intent_name, self.confidence_score)
        

# See: https://docs.snips.ai/reference/dialogue#intent
class IntentPayload:
    
    def __init__(self, json_dict):
        self.json = json_dict
        
        self.session_id = json_dict['sessionId']
        self.input = json_dict['input']
        self.intent = Intent(json_dict['intent'])
        self.site_id = json_dict['siteId']

        custom_data = json_dict.get('customData')
        try:
            self.custom_data = json.loads(custom_data)
        except:
            self.custom_data = custom_data
        
        self.asr_tokens = json_dict.get('asrTokens')
        self.asr_confidence = json_dict.get('asrConfidence')
        self.slots = { s['slotName'] : Slot(i, s)
            for i, s in enumerate(json_dict.get('slots', [])) }
        self.slot_values = { name : s.value
            for name, s in self.slots.items() }

    def __repr__(self):
        return "<%s %s>" % (self.intent, list(self.slots))



from . exceptions import SnipsError, SnipsClarificationError
from . i18n import *


__all__ = ('MultiRoomConfig', 'ROOMS', 'room_with_article', 'room_with_preposition')

_, ngettext = get_translations(__file__, 'snips_skill')


class MultiRoomConfig:
    ''' Mixin for multi-site actions.
        Contains helper methods to extract location slots,
        look up room names, associated configuration and site IDs.
    '''
    
    LOCATION_SLOT = None # Override as needed, e.g. 'room'


    def process_config(self):
        'Load all non-standard config sections as rooms'
        self.sites = { self.configuration[section]['site_id'] : section
            for section in self.configuration
            if section not in self.STANDARD_SECTIONS
                and 'site_id' in self.configuration[section] }


    def add_room_name(self, room, with_article, with_preposition):
        'Register additional room names with articles and prepositions'
        ROOMS[room.lower()] = RoomName(with_article, with_preposition)


    def get_current_room(self, payload):
        'Get the room name of the current site'
        return self.sites.get(payload.site_id)

        
    def get_room(self, payload):
        'Get the recognized room name'
        
        default_room = room = self.get_current_room(payload)
        assert self.LOCATION_SLOT, 'self.LOCATION_SLOT is undefined'
        if self.LOCATION_SLOT in payload.slot_values:
            room = payload.slot_values[self.LOCATION_SLOT].value
            if default_room and room in DEFAULT_ROOM_NAMES:
                return default_room
        return room
                

    def in_current_room(self, payload):
        'Is the current site the disired room?'
        return self.get_room(payload) == self.get_current_room(payload)

        
    def get_room_name(self, payload, modifier=str, default=None):
        ''' Get the recognized room name,
            optionally adding an article or preposition
        '''
        
        room = self.get_room(payload)
        default_room = self.get_current_room(payload)
        if room and default is not None and room == default_room:
            return default
        return modifier(room or _('unknown room'))


    def get_room_config(self, payload):
        ''' Get the configuration section for a recognized room.
            :param payload: parsed intent message payload
            :return: room configuration
        '''

        room = self.get_room(payload)
        if room in self.configuration: return self.configuration[room]
        raise SnipsClarificationError(_('in which room?'),
            payload.intent.intent_name, self.LOCATION_SLOT)

    
    def all_rooms(self, payload):
        'Check whether the intent refers to every room at once'
        room = self.get_room(payload).lower()
        for name in ALL_ROOMS:
            name = name.lower()
            if name == room or name in payload.input:
                return True
        

    def get_site_id(self, payload):
        ''' Obtain a site_id by explicit or implied room name.
            :param payload: parsed intent message payload
        '''
        
        config = self.get_room_config(payload)
        if 'site_id' in config: return config['site_id']
        raise SnipsError(_('This room has not been configured yet.'))



#!/usr/bin/env python3

'''
    Simplistic wrapper for the Paho MQTT client.
'''

from argparse import ArgumentParser
from functools import wraps
from getpass import getpass
from paho.mqtt.client import Client as PahoClient, MQTTv311
import json, logging


__all__ = ('MqttClient', 'CommandLineMixin', 'topic', 'CommandLineClient')


class MqttClient(PahoClient):
    'MQTT client'
    
    TCP = 'tcp'
    WEBSOCKETS = 'websockets'
    DEFAULT_PORT = 1883
    DEFAULT_TLS_PORT = 8883
    SUBSCRIPTIONS = {}
    

    def __init__(self, client_id=None, clean_session=True, \
        userdata=None, protocol=MQTTv311, transport=TCP):

        super().__init__(client_id, clean_session or not client_id,
            userdata, protocol, transport)
        self._tls_initialized = False
        self.log = logging.getLogger(self.__class__.__name__)


    def __enter__(self):
        return self


    def __exit__(self, *exc_info):
        self.disconnect()


    def connect(self, host='localhost', port=DEFAULT_PORT,
            username=None, password=None,
            keepalive=60, bind_address='', use_tls=False):
        'Connect to the MQTT broker'
        
        if username: self.username_pw_set(username, password)
        if use_tls or port == self.DEFAULT_TLS_PORT:
            if not self._tls_initialized: self.tls_set()
        self.log.debug("Connecting to MQTT broker %s as user '%s'",
            host, username or '')
        super().connect(host, port, keepalive, bind_address)
        return self


    def disconnect(self):
        super().disconnect()
        self.log.debug('Disconnected from MQTT broker')


    def reconnect(self):
        self.log.debug('Reconnecting to MQTT broker')
        super().reconnect()
    
        
    def loop_forever(self):
        'Wait for messages and invoke callbacks until interrupted'
        try:
            super().loop_forever()
            
        except KeyboardInterrupt:
            self.log.info('Interrupted by user')
    
    
    def subscribe(self, topic, qos=0):
        'Subscribe to a MQTT topic'
        super().subscribe(topic, qos)
        self.log.debug('Subscribed to MQTT topic: %s', topic)
        
        
    def publish(self, topic, payload=None, qos=0, retain=False):
        'Send an MQTT message'
        self.log.debug('Publishing: %s', topic)
        return super().publish(topic, payload, qos, retain)


    def on_connect(self, client, userdata, flags, rc):
        'Subscribe to MQTT topics'
        
        assert rc == 0, 'Connection failed'
        assert self == client, 'Bad karma'
        self.log.debug('Connected to MQTT broker')
        
        # Register @topic callbacks
        for topic, (callback, qos) in self.SUBSCRIPTIONS.items():
            self.subscribe(topic, qos)
            self.message_callback_add(topic, callback)


    def decode_json(self, payload):
        'Try to decode a message payload as JSON'
        try: return json.loads(payload)
        except ValueError: return payload
    


def topic(topic, qos=0, payload_converter=None, log_level=logging.DEBUG):
    ''' Decorator for callback functions.
        Callbacks are invoked with these positional parameters:
         - client: MqttClient instance
         - msg: MQTT message
         - userdata: User-defined extra data
        Return values are not expected.
        :param topic: MQTT topic, may contain wildcards
        :param qos: MQTT quality of service (default: 0)
        :param payload_converter: unary function to transform the message payload
    '''
    
    assert topic not in MqttClient.SUBSCRIPTIONS, \
        "Topic '%s' is already registered" % topic
    
    def wrapper(method):
        
        @wraps(method)
        def wrapped(client, userdata, msg):
            'Callback for the Paho MQTT client'
            if log_level: 
                client.log.log(log_level, 'Received message: %s', msg.topic)
            if payload_converter: msg.payload = payload_converter(msg.payload)
            
            # User-provided callback
            return method(client, userdata, msg)

        MqttClient.SUBSCRIPTIONS[topic] = (wrapped, qos)
        return wrapped
    return wrapper
  

class CommandLineMixin:
    'Provides logging and standard command-limne arguments'

    # Map verbosity argument choices to log levels
    LOG_LEVELS = {
        0: logging.ERROR,
        1: logging.WARNING,
        2: logging.INFO,
        3: logging.DEBUG,
    }
    
    DEFAULT_LOG_LEVEL = 2
    LOG_FORMAT = '%(message)s'


    def __init__(self):
        super().__init__()
        self.parse_args()


    def parse_args(self, args=None):
        logging.basicConfig(format=self.LOG_FORMAT)
        self.parser = ArgumentParser(description=self.__doc__)
        self.parser.add_argument('-v', '--verbosity',
            type=int, choices=[0, 1, 2, 3], default=self.DEFAULT_LOG_LEVEL,
            help='Logging verbosity: 0=errors only, 1=errors and warnings, 2=normal output, 3=debug output')
        self.add_arguments()
        
        self.options = self.parser.parse_args(args)
        self.log.setLevel(self.LOG_LEVELS[self.options.verbosity])
        self.log.debug('Command line options: %s', self.options)
    

    def add_arguments(self):
        'Hook for subclasses to add additional command line options'
        pass
 
 
class CommandLineClient(CommandLineMixin, MqttClient):
    
    'Simple MQTT command line client'

    password = None
    
    
    def add_arguments(self):
        'Set up arguments for connection parameters'        
        self.parser.add_argument('-H', '--host', default='localhost',
            help='MQTT host (default: localhost)')
        self.parser.add_argument('-P', '--port', default=MqttClient.DEFAULT_PORT,
            type=int, help='MQTT port (default: %d)' % MqttClient.DEFAULT_PORT)
        self.parser.add_argument('-T', '--tls', action='store_true',
            default=False, help='Use TLS')
        self.parser.add_argument('-u', '--username', nargs='?', help='User name')
        self.parser.add_argument('-p', '--password', action='store_true',
            help='Prompt for password')
    
    
    def parse_args(self):
        super().parse_args()
        if self.options.username and self.options.password:
            self.password = getpass()
        if self.options.tls and self.options.port == MqttClient.DEFAULT_PORT:
            self.options.port = MqttClient.DEFAULT_TLS_PORT
    
    
    def run(self):
        'Connect to MQTT and handle incoming messages'
        with self.connect(self.options.host, self.options.port,
            self.options.username, self.password, use_tls=self.options.tls):
            self.loop_forever()
        
        
if __name__ == '__main__': # Demo code

    from colors import cyan
    from shutil import get_terminal_size
    import sys
    
    class Logger(CommandLineClient):
                
        WIDTH = get_terminal_size().columns
        COLOR = cyan if sys.stdout.isatty() else str
        
        def add_arguments(self):
            super().add_arguments()
            self.parser.add_argument('-t', '--topic', default='#',
                help='MQTT topic (default: #)')
            self.parser.add_argument('-w', '--width', default=self.WIDTH,
                type=int, help='Output width (default: %d)' % self.WIDTH)
            self.parser.add_argument('-j', '--json', action='store_true',
                help='Try to decode JSON payloads')
            self.parser.add_argument('-z', '--clear', action='store_true',
                help='Clear retained messages')

    client = Logger()

    @topic(client.options.topic)
    def print_msg(client, userdata, msg):
        converter = client.decode_json if client.options.json else str
        width = client.options.width - len(msg.topic) - 2
        payload = converter(msg.payload)
        client.log.info('%s: %.*a', client.COLOR(msg.topic), width, payload)
        if client.options.clear and msg.retain and msg.payload:
            client.publish(msg.topic, retain=True)

    client.run()



from collections import namedtuple
import gettext, locale, os


__all__ = ('_', 'ALL_ROOMS', 'CONFIRMATIONS', 'DEFAULT_ROOM_NAMES', 'ROOMS', 
    'get_translations', 'ngettext', 'room_with_article', 'room_with_preposition')


def get_translations(file, domain='messages'):
    'Install translations'
    language, encoding = locale.getlocale()
    locale.setlocale(locale.LC_ALL, (language, encoding))
    translation = gettext.translation(domain, 
        languages=[language], fallback=True,
        localedir=os.path.join(os.path.dirname(file), 'locale'))
    return translation.gettext, translation.ngettext


_, ngettext = get_translations(__file__, 'snips_skill')

RoomName = namedtuple('RoomName', 'with_article, with_preposition')

CONFIRMATIONS = (
    _('done'),
    _('here you are'),
    _('OK'),
    _('yes'),
    _('finished'),
    _('alright'),
)

DEFAULT_ROOM_NAMES = (
    _('here'),
    _('this room'),
)

ALL_ROOMS = (
    _('all'),
    _('everywhere'), 
    _('everything'),
)

ROOMS = {
    # Map translated room names to room names with articles and prepositions
    # for languages that use genders for room names, e.g. German
    _("bathroom").lower():     RoomName(_("the bathroom"),    _("in the bathroom")),
    _("bedroom").lower():      RoomName(_("the bedroom"),     _("in the bedroom")),
    _("dining room").lower():  RoomName(_("the dining room"), _("in the dining room")),
    _("livingroom").lower():   RoomName(_("the livingroom"),  _("in the livingroom")),
    _("kid's room").lower():   RoomName(_("the kid's room"),  _("in the kid's room")),
    _("kitchen").lower():      RoomName(_("the kitchen"),     _("in the kitchen")),
    _("office").lower():       RoomName(_("the office"),      _("in the office")),
    _("hall").lower():         RoomName(_("the hall"),        _("in the hall")),
    _("garden").lower():       RoomName(_("the garden"),      _("in the garden")),
    _("unknown room").lower(): RoomName(_("an unknown room"), _("in an unknown room")),
}

        
def room_with_article(room_name):
    'Get the spoken room name with the definite article'
    room = ROOMS.get(room_name.lower())
    return room.with_article if room \
        else _('the {room}').format(room=room_name)
    

def room_with_preposition(room_name):
    'Get the spoken room name with "in" preposition'
    room = ROOMS.get(room_name.lower())
    return room.with_preposition if room \
        else _('in the {room}').format(room=room_name)



from argparse import ArgumentTypeError, FileType
from colors import red, green, yellow, blue, magenta, cyan
from datetime import datetime
from pathlib import Path
import json, logging

from . logging import LoggingMixin
from . mqtt import *
from . snips import *


class TestRunner(CommandLineMixin, LoggingMixin, SnipsClient):
    'Record Snips sessions and play them back'

    def __init__(self):
        super().__init__()
        
        self.test = None
        self.events = []
        self.failures = 0
        

    def add_arguments(self):
        super().add_arguments()
        self.parser.add_argument('-d', '--log-dir',
            type=Path, nargs='?', help='Directory to log JSON messages')
        self.parser.add_argument('-i', '--ignore-text', action='store_true',
            help='Do not compare TTS message texts')
        self.parser.add_argument('-l', '--loop', action='store_true',
            help='Keep recording until interrupted')
        self.parser.add_argument('tests', nargs='*', type=self._json_file,
            metavar='JSON_TEST', help='JSON test spec')
    
         
    def _json_file(self, path):
        if Path(path).suffix != '.json':
            raise ArgumentTypeError('%s is not a JSON file' % path) 
        return FileType('r')(path)
    
        
    def on_connect(self, client, userdata, flags, rc):
        super().on_connect(client, userdata, flags, rc)
        
        if self.options.log_dir: # Start recording
            if not self.options.log_dir.is_dir():
                self.log.error('No such directory: %s', self.options.log_dir)
                return self._exit()
            self.log.info('Waiting for session...')
        
        elif self.options.tests: self._start_session()
        else: self._exit('Nothing to do, exiting')

    
    def _start_session(self):
        with self.options.tests.pop(0) as test:
            self.test = json.load(test)
        site_id = self.test[0].get('payload', {}).get('siteId')
        self.start_session(site_id, self.action_init())


    @on_session_started()
    def _on_start(self, userdata, msg):
        if self.options.log_dir: return # Recording

        try:
            self.session_id = msg.payload.get('sessionId')
            self.log.debug('Session started: %s', self.session_id)
    
            assert self.test, 'No test steps remaining'
            step = self.test[0]

            topic = step.get('topic')
            assert topic, 'Message topic is missing'
            assert topic.startswith(self.INTENT_PREFIX), \
                'Intent expected, but got: %s' % topic

            payload = step.get('payload')
            assert payload, 'Message payload is missing'
            payload['sessionId'] = self.session_id

            self.log.info('Starting test %s with %d steps',
                step.get('time'), len(self.test))
            self.publish(topic, json.dumps(payload))
            return
            
        except AssertionError as e:
            self._fail(e)    
        self._exit()


    @on_intent('#', log_level=None)
    @on_continue_session(log_level=None)
    @on_end_session(log_level=None)
    def _handle(self, userdata, msg):
        self.log.debug('Received message: %s', msg.topic)
        if self.options.log_dir:
            self.events.append({
                'time': datetime.now().isoformat(),
                'topic': msg.topic,
                'payload': msg.payload })
            return

        # Ignore other sessions that are unrelated to the current test
        if msg.payload and msg.payload.get('sessionId') != self.session_id:
            return

        try:
            assert self.test, 'Test has already finished'
            step = self.test.pop(0)
            topic = step.get('topic')
            assert topic, 'Test topic is missing'
            self.log.debug('Test step: %s', topic)
            
            assert topic == msg.topic, \
                'Expected topic: %s, received: %s' % (topic, msg.topic)

            payload = step.get('payload')
            assert payload, 'Test payload is missing'
            
            text = msg.payload.get('text')
            assert (self.options.ignore_text or payload.get('text') == text), \
                'Expected text: %s, received: %s' % (payload.get('text'), text)

            if topic.startswith(self.INTENT_PREFIX):
                self.tabular_log(logging.INFO, 'intent', topic, label_color=green)
            else: self.tabular_log(logging.INFO, 'topic', topic, label_color=cyan)
            if text: self.tabular_log(logging.INFO, 'text', text, label_color=cyan)
            
        except AssertionError as e: self._fail(e)


    @on_session_ended()
    def _on_end(self, userdata, msg):
        if self.test:
            self._fail('Test has %d remaining steps' % len(self.test))
        self.log.debug('Session ended: %s', msg.payload['sessionId'])
        if self.options.tests: self._start_session()
        else: self._exit()


    def _fail(self, e):
        self.colored_log(logging.ERROR, '%s', e, color=red)
        self.failures += 1
        self.test = None


    def _exit(self, msg='Exiting'):
        if self.events:
            event = self.events[0]
            intent = event['topic'].replace(self.INTENT_PREFIX, '')
            file_name = '%s-%s.json' % (event['time'], intent)
            path = self.options.log_dir / file_name
            self.log.info('Logging session to %s', path)
            with open(path, 'w') as out:
                json.dump(self.events, out, ensure_ascii=False, indent=2)
            self.events = []
            
        if not (self.options.log_dir and self.options.loop):
            self.log.debug(msg)
            self.disconnect()


if __name__ == '__main__':
    import sys

    client = TestRunner()
    client.run()
    sys.exit(client.failures)



from . exceptions import *
from . i18n import get_translations, CONFIRMATIONS
from . logging import *
from . mqtt import *
from . multi_room import *
from . snips import *
from . skill import *
from . state import *


__all__ = ('CommandLineClient', 'CommandLineMixin', 'LoggingMixin', 'MqttClient',
    'Skill', 'SnipsClient', 'SnipsError', 'SnipsClarificationError', 'StateAwareMixin',
    'debug_json', 'get_translations', 'topic', 'min_confidence', 'MultiRoomConfig',
    'on_intent', 'intent', 'PARDON', 'require_slot',
    'on_hotword_detected', 'on_start_session', 'on_session_started',
    'on_end_session', 'on_continue_session', 'on_session_ended', 'on_play_finished',
    'CONFIRMATIONS', 'ROOMS', 'room_with_article', 'room_with_preposition')

__version__ = '0.1.2'
